"""Utility to hide imports during tests."""

import sys
from contextlib import contextmanager
from types import ModuleType
from typing import Dict


@contextmanager
def without_modules(*modules: str):
    """Run the code inside the context manager without the specified modules.

    Args:
        modules (*str): The names of the modules to hide from import.

    Yields:
        None
    """
    # Get all currently loaded modules
    excluded_modules = [m for m in sys.modules.keys() if any(m.startswith(ex_m) for ex_m in modules)]

    extracted_modules: Dict[str, ModuleType] = {}

    # Delete them from the sys.modules
    # Setting them to None triggers an ImportError on attempted
    # import
    for m in excluded_modules:
        extracted_modules[m] = sys.modules[m]
        sys.modules[m] = None  # type: ignore

    yield

    # Clear the None so we can re-import if necessary
    for n in excluded_modules:
        sys.modules[n] = extracted_modules[n]



"""These functions only exist to test the dynamic coverage."""


def fn_only_covered_in_unit():  # pragma: only-covered-in-unit-tests
    return True


def fn_only_covered_in_integration():  # pragma: only-covered-in-integration-tests
    return True


# This line is here to ensure that there is always 1 line of coverage for the file
marker = True



from invoke import Collection
from outcome.utils import env

from outcome.devkit.invoke.tasks import check, clean, database, pact, release, setup, test

namespace = Collection(setup, clean, release, check, test, database, pact)

namespace.configure({'run': {'echo': True, 'pty': not env.is_ci()}})



from typing import cast

from invoke import run

from outcome.devkit.invoke import env


@env.add
def git_branch(e: env.Env):
    return cast(str, run('git rev-parse --abbrev-ref HEAD').stdout.strip())


@env.add
def commit_sha1(e: env.Env):
    return cast(str, run('git rev-parse --short=8 HEAD').stdout.strip())



"""Docker helpers."""

from enum import Enum
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, cast

from invoke import Context, Result
from pydantic import BaseModel, Field, validator

EnvironmentVar = Dict[str, str]
CommandArgs = Sequence[Tuple[str, ...]]
Volumes = Sequence[Tuple[str, str]]


class UnknownContainer(Exception):
    ...


class DuplicateContainer(Exception):
    ...


default_args = ['--rm']


def create_container(  # noqa: WPS211, WPS231
    c: Context,
    name: str,
    image: str,
    environment: Optional[EnvironmentVar] = None,
    port: Optional[int] = None,
    workdir: Optional[str] = None,
    volumes: Optional[Volumes] = None,
    detach: bool = True,
    additional_args: Optional[List[str]] = None,
    command: Optional[str] = None,
    command_args: Optional[CommandArgs] = None,
    command_args_sep: str = ' ',
):
    environment = environment or {}
    volumes = volumes or []
    additional_args = additional_args or []

    if container_exists(c, name):
        raise DuplicateContainer(name)

    args = default_args.copy()

    args.append(f'--name {name}')

    if detach:
        args.append('-d')
    else:
        args.append('-it')

    if workdir:
        args.append(f'-w {workdir}')

    for volume in volumes:
        args.append(f'-v {volume[0]}:{volume[1]}')

    if port:
        args.append(f'-p {port}:{port}')

    for key, value in environment.items():
        args.append(f'-e {key}="{value}"')

    arg_string = ' '.join(args + additional_args)

    command_to_run = f'docker run {arg_string} {image}'

    if command:
        command_to_run = f'{command_to_run} {command}'

        if command_args:
            for command_arg in command_args:
                command_arg_string = command_args_sep.join(command_arg)
                command_to_run = f'{command_to_run} {command_arg_string}'

    c.run(command_to_run)


def start_container(c: Context, name: str):
    if not container_exists(c, name):
        raise UnknownContainer

    if not container_is_running(c, name):
        c.run(f'docker start {name}')


def stop_container(c: Context, name: str):
    if container_is_running(c, name):
        c.run(f'docker stop {name}')


class DockerState(Enum):
    running = 'running'
    exited = 'exited'


class DockerProcess(BaseModel):
    class Config:
        extra = 'ignore'

    command: str = Field(..., alias='Command')
    identifier: str = Field(..., alias='ID')
    image: str = Field(..., alias='Image')
    state: DockerState = Field(..., alias='State')
    names: List[str] = Field(..., alias='Names')

    @validator('names', pre=True)
    def parse_list(cls, v: object) -> object:  # noqa: N805  # pragma: no cover
        if isinstance(v, str):
            return v.split(',')
        return v


def container_exists(c: Context, name: str) -> bool:
    for container in get_docker_containers(c, only_running=False):
        if name in container.names:
            return True

    return False


def container_is_running(c: Context, name: str) -> bool:
    for container in get_docker_containers(c):
        if name in container.names:
            return container.state == DockerState.running

    return False


def get_docker_containers(c: Context, only_running: bool = True) -> Iterator[DockerProcess]:
    options: List[str] = ["--format '{{json .}}'"]

    if only_running:
        options.append('-a')

    opts = ' '.join(options)
    command = f'docker ps {opts}'

    res = cast(Result, c.run(command, echo=False, hide=True))
    output = cast(str, res.stdout)

    yield from (DockerProcess.parse_raw(line) for line in output.splitlines() if line)



from pathlib import Path

from outcome.read_toml import lib as read_toml
from outcome.utils.config import Config

from outcome.devkit.invoke import env

pyproject_path = Path.cwd() / 'pyproject.toml'

pyproject_file = env.declare('pyproject.toml', str(pyproject_path))
config = Config(pyproject_path)


@env.add
def package_name(e: env.Env) -> str:
    return read_toml.read_from_file(env.r(pyproject_file), 'tool.poetry.name')


@env.add
def package_version(e: env.Env) -> str:
    return read_toml.read_from_file(env.r(pyproject_file), 'tool.poetry.version')



from typing import Optional

from invoke import Collection, Context, task
from outcome.read_toml.lib import read_from_file

from outcome.devkit.invoke import env
from outcome.devkit.invoke.tasks import clean


def get_key_from_pyproject(e: env.Env, key: str):
    try:
        pyproject = e.read('pyproject.toml')
        assert pyproject
        return read_from_file(pyproject, key)
    except KeyError:
        return None


@env.add(required=False)
def package_repository_name(e: env.Env):
    return get_key_from_pyproject(e, 'package.repository.name')


@env.add(required=False)
def package_repository_url(e: env.Env):
    return get_key_from_pyproject(e, 'package.repository.url')


@task(clean.all)
def build(c: Context):
    """Build the python package."""
    c.run('poetry build')


@task(build)
def publish(c: Context, repository_name: Optional[str] = None, repository_url: Optional[str] = None):
    """Publish the python package."""

    if any((repository_name, repository_url)) and not all((repository_name, repository_url)):
        raise RuntimeError('You must provide both name and URL when specifying a custom package repository')

    repo_name = repository_name or env.nullable_read(package_repository_name)
    repo_url = repository_url or env.nullable_read(package_repository_url)

    if repo_name and repo_url:
        c.run(f'poetry config repositories.{repo_name} {repo_url}')
        c.run(f'poetry publish -n -r {repo_name}')
    else:
        c.run('poetry publish -n')


ns = Collection(build, publish)



from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Literal, Optional, Sequence, Set, cast

from invoke import Collection, Context, Result, UnexpectedExit, task
from outcome.utils.env import is_ci
from pydantic import BaseModel, Field
from rich.console import Console

from outcome.devkit.invoke import env
from outcome.devkit.invoke.tasks import clean as clean_tasks


class PyrightDiagnosticRangeBoundary(BaseModel):
    line: int
    character: int


class PyrightDiagnosticRange(BaseModel):
    start: PyrightDiagnosticRangeBoundary
    end: PyrightDiagnosticRangeBoundary


class PyrightDiagnostic(BaseModel):
    file: str
    severity: Literal['error', 'warning', 'information']
    message: str
    rule: Optional[str]
    range: PyrightDiagnosticRange  # noqa: A003, WPS125


class PyrightOutputSummary(BaseModel):
    error_count: int = Field(..., alias='errorCount')
    files_analyzed: int = Field(..., alias='filesAnalyzed')
    warning_count: int = Field(..., alias='warningCount')


class PyrightOutput(BaseModel):
    class Config:
        extra = 'ignore'

    general_diagnostics: Sequence[PyrightDiagnostic] = Field(..., alias='generalDiagnostics')
    summary: PyrightOutputSummary


@env.add
def check_targets(e: env.Env) -> str:
    return ' '.join(d for d in ('./src', './bin', './test', './tasks.py') if Path(d).exists())


@task(clean_tasks.docs, clean_tasks.python, clean_tasks.coverage)
def clean(c: Context):  # noqa: A001, WPS125
    """Clean everything for checks."""
    ...


@task(clean)
def types(c: Context, show_information: bool = False):
    """Run type-checking."""
    directories = env.r(check_targets)

    tmp_handler = NamedTemporaryFile()
    res = cast(
        Result, c.run(f'poetry run pyright {directories} --outputjson > {tmp_handler.name}', echo=True, hide=True, warn=True),
    )

    results = PyrightOutput.parse_raw(tmp_handler.read())

    console = Console()

    files_analyzed = results.summary.files_analyzed
    warning_count = results.summary.warning_count
    error_count = results.summary.error_count

    console.print(
        f'pyright analyzed {files_analyzed} file(s), with {warning_count} warnings and {error_count} errors',  # noqa: E501
        style='cyan',
    )

    skipped_statuses: Set[str] = set()
    if not show_information:
        skipped_statuses.add('information')

    for d in (d for d in results.general_diagnostics if d.severity not in skipped_statuses):  # noqa: WPS335
        if d.severity == 'error':
            color = 'red'
        elif d.severity == 'warning':
            color = 'yellow'
        else:
            color = 'blue'

        message = f'[dim]({d.rule})[/dim] {d.message}' if d.rule else d.message
        file = d.file
        line = d.range.start.line + 1
        col = d.range.start.character + 1
        severity = d.severity

        console.print(f'{file}:{line}:{col} - [bold {color}]{severity}[/bold {color}] {message}')

    if results.summary.error_count > 0:
        res.hide = ()
        raise UnexpectedExit(res, 'Error checking types')


@task(clean)
def format(c: Context):  # noqa: A001, WPS125
    """Run formatter."""
    directories = env.r(check_targets)
    if is_ci():
        c.run(f'poetry run black --check {directories}')
    else:
        c.run(f'poetry run black {directories}')


@task(clean)
def isort(c: Context):
    """Run isort."""
    directories = env.r(check_targets)
    if is_ci():
        c.run(f'poetry run isort {directories} --check-only')
    else:
        c.run(f'poetry run isort {directories}')


@task(clean)
def lint(c: Context):
    """Run flake8."""
    directories = env.r(check_targets)
    c.run(f'poetry run flake8 {directories}')


@task(clean, types, isort, format, lint)
def all(c: Context):  # noqa: A001, WPS125
    """Run all checks."""
    ...


ns = Collection(lint, isort, format, types)
ns.add_task(all, default=True)



import json
import os
from pathlib import Path

from invoke import Collection, Context, task

from outcome.devkit.invoke import docker, env
from outcome.devkit.invoke.tasks import git, package

pact_broker_url = env.declare('PACT_BROKER_URL', 'https://pact.svc.outcome.co')
pact_providers_dir = env.declare('PACT_PROVIDERS_DIR', 'pacts')
pact_foundation_cli_image = env.declare('PACT_FOUNDATION_CLI_IMAGE', 'pactfoundation/pact-cli:latest')

pact_broker_username = env.from_os('PACT_BROKER_USERNAME')
pact_broker_password = env.from_os('PACT_BROKER_PASSWORD')
pact_github_token = env.from_os('PACT_GITHUB_TOKEN')


def pact_foundation_vars():
    return {
        'PACT_BROKER_BASE_URL': env.r(pact_broker_url),
        'PACT_BROKER_USERNAME': env.r(pact_broker_username),
        'PACT_BROKER_PASSWORD': env.r(pact_broker_password),
    }


event_pact_changed = {
    'event_type': 'pact_content_changed',
    'client_payload': {'pact_url': '${pactbroker.pactUrl}', 'consumer': '${pactbroker.consumerName}'},
}


@task
def create_webhooks_from_pacts(c: Context):  # noqa: A001, WPS125
    """Create webhooks in Pact Broker from local pact files."""
    for file in Path(env.r(pact_providers_dir)).glob('*'):
        with open(file, 'r') as handler:
            json_file = json.load(handler)
        provider_name = json_file['provider']['name']
        create_webhooks_for_provider(c, provider_name)


@task
def create_webhooks_for_provider(c: Context, provider_name: str):  # noqa: A001, WPS125
    """Create webhook in Pact Broker for a provider."""
    var_pact_github_token = env.r(pact_github_token)
    var_event_pact_changed = json.dumps(event_pact_changed)

    docker.create_container(
        c,
        f'webhook-{provider_name}',
        image=env.r(pact_foundation_cli_image),
        environment=pact_foundation_vars(),
        command='broker',
        command_args=[
            ('create-webhook', f'https://api.github.com/repos/outcome-co/{provider_name}/dispatches'),
            ('--contract-content-changed',),
            ('--provider', provider_name),
            ('-X', 'POST'),
            ('-H', 'User-Agent: PactBroker'),
            ('-H', 'Host: api.github.com'),
            ('-H', 'Content-Type: application/json'),
            ('-H', 'Accept: application/vnd.github.v3+json'),
            ('-H', f'Authorization: token ${var_pact_github_token}'),
            ('-d', f"'{var_event_pact_changed}'"),
        ],
    )


@task
def publish_as_consumer(c: Context):  # noqa: A001, WPS125
    """Publish local Pact files to Pact Broker."""
    var_package_version = env.r(package.package_version)
    var_pact_providers_dir = env.r(pact_providers_dir)
    var_cwd = os.getcwd()

    docker.create_container(
        c,
        f'publish-{var_package_version}',
        image=env.r(pact_foundation_cli_image),
        environment=pact_foundation_vars(),
        volumes=[(var_cwd, '/app')],
        workdir='/app',
        command='publish',
        command_args=[
            (f'/app/{var_pact_providers_dir}',),
            ('--consumer-app-version', env.r(git.commit_sha1)),
            ('--tag', env.r(package.package_version)),
            ('--tag', env.r(git.git_branch)),
        ],
    )


@task
def create_version_tags(c: Context):  # noqa: A001, WPS125
    """Tag Pact with the version of the package."""
    var_package_version = env.r(package.package_version)
    var_cwd = os.getcwd()

    docker.create_container(
        c,
        f'tags-{var_package_version}',
        image=env.r(pact_foundation_cli_image),
        environment=pact_foundation_vars(),
        volumes=[(var_cwd, '/app')],
        workdir='/app',
        command='broker',
        command_args=[
            ('create-version-tag',),
            ('--pacticipant', env.r(package.package_name)),
            ('--version', env.r(git.commit_sha1)),
            ('--tag', env.r(package.package_version)),
            ('--tag', env.r(git.git_branch)),
        ],
    )


ns = Collection(create_webhooks_from_pacts, create_webhooks_for_provider, publish_as_consumer, create_version_tags)



from invoke import Collection, Context, task
from outcome.read_toml import lib as read_toml
from outcome.utils.config import Config

from outcome.devkit.invoke import docker, env
from outcome.devkit.invoke.tasks import package

defaults = {
    'DB_IMAGE': 'postgres:latest',
    'DB_NAME': 'postgres',
    'DB_USER': 'postgres',
    'DB_PASSWORD': 'postgres',
    'DB_HOST': '127.0.0.1',
    'DB_PORT': '5432',
}


config = Config(package.pyproject_path, defaults=defaults)


def read_or_default(key: str, default: str) -> str:
    try:
        return read_toml.read_from_file(env.r(package.pyproject_file), key)
    except KeyError:
        return default


database_image = env.from_config('DB_IMAGE', config=config)
database_name = env.from_config('DB_NAME', config=config)
database_user = env.from_config('DB_USER', config=config)
database_password = env.from_config('DB_PASSWORD', config=config)
database_port = env.from_config('DB_PORT', config=config)
database_host = env.from_config('DB_HOST', config=config)


@env.add
def database_container(e: env.Env) -> str:
    name = env.r(package.package_name)
    return f'{name}-db'


@task
def start(c: Context, container_name: str = env.r(database_container)):
    """Start the database container."""
    # If the container is running, there's nothing to do
    if docker.container_is_running(c, container_name):
        return

    if docker.container_exists(c, container_name):
        docker.start_container(c, container_name)

    container_image = env.r(database_image)
    container_port = int(env.r(database_port))
    container_password = env.r(database_password)

    container_env = {'POSTGRES_PASSWORD': container_password}

    # Container doesn't exist, we need to create it
    docker.create_container(c, container_name, image=container_image, port=container_port, environment=container_env)


@task
def stop(c: Context, container_name: str = env.r(database_container)):
    """Stop the database container."""
    if docker.container_is_running(c, container_name):
        docker.stop_container(c, container_name)


ns = Collection(start, stop)
